#!/usr/bin/env python

"""
url: https://github.com/zsimic
download_url: archive/v{version}.tar.gz
"""

from setuptools import setup


setup(
    name="runez",
    setup_requires="setupmeta",
    versioning="dev",
    author="Zoran Simic zoran@simicweb.com",
)
